--
-- PIC16F636 compatible microcontroller core
--
-- Version : 0.1.02
-- 
-- Copyright (c) 2010 Yannick Bornat (yannick.bornat@enseirb-matmeca.fr)
--
-- All rights reserved
--
-- Redistribution and use in source and synthezised forms, with or without
-- modification, are permitted provided that the following conditions are met:
--
-- Redistributions of source code must retain the above copyright notice,
-- this list of conditions and the following disclaimer.
--
-- Redistributions in synthesized form must reproduce the above copyright
-- notice, this list of conditions and the following disclaimer in the
-- documentation and/or other materials provided with the distribution.
--
-- Neither the name of the author nor the names of other contributors may
-- be used to endorse or promote products derived from this software without
-- specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
-- THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-- PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.
--
-- Please report bugs to the author, but before you do so, please
-- make sure that this is not a derivative work and that
-- you have the latest version of this file.
--
--
-- Limitations :
--	No PortA PortC
-- No Timer 1
-- no WDT
--	No EEPROM 
--

-- Known bugs :
--  writing to PCL has no effect
-- 

LIBRARY IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.numeric_std.ALL;

ENTITY OLED_CPU IS
  PORT(
    Clk           : IN std_logic;
    reset_n       : IN std_logic;
    
    ext_data_in   : in  std_logic_vector(7 DOWNTO 0);           --
    ext_data_out  : out std_logic_vector(7 DOWNTO 0);           -- 
    ext_data_addr : out std_logic_vector(8 DOWNTO 0);           -- 
    ext_we        : out std_logic;                              -- 
    ext_re        : out std_logic;                              -- 
    
    code_addr     : out std_logic_vector(10 downto 0);
    code_data     : in std_logic_vector(13 downto 0)
    );
END OLED_CPU;

ARCHITECTURE basic OF OLED_CPU IS

TYPE RAM_BBLOCK IS ARRAY (127 DOWNTO 0) OF std_logic_vector(7 DOWNTO 0);
TYPE t_cycle is (Dinstr, Dread, Dcomp, Dwrite, Dsleep);
TYPE t_instr_code is (movwf, clrn, subn, decn, iorn, andn, xorn, addn, movn, comf, incn, rrf, rlf, swapf, bcf);

-- constants
constant wdt_timing : std_logic_vector(19 downto 0):="11011011101110011111"; -- 18ms @ 50MHz

-- synchronous signals
signal bRAM         : RAM_BBLOCK:=(others=>"00000000");                     -- the 128 bytes embedded RAM
signal cycle        : t_cycle;                        -- processor machine cycle
signal instr_code   : t_instr_code;                   -- used for instruction decoding
signal w_reg        : std_logic_vector(7 downto 0):="01010000";   -- W register
signal Instruction  : std_logic_vector(13 downto 0):=(others=>'0');  -- curent instruction
signal carry        : std_logic :='0' ;                      -- carry bit
signal digcarry     : std_logic:='0';                      -- digit carry bit
signal zero         : std_logic:='0';                      -- zero bit
signal pc           : std_logic_vector(10 downto 0):=(others=>'0');   -- program counter, only 10 bits for 16F84
signal stack0       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 0 (younger)
signal stack1       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 1 
signal stack2       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 2 
signal stack3       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 3 
signal stack4       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 4 
signal stack5       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 5 
signal stack6       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 6 
signal stack7       : std_logic_vector(10 downto 0):=(others=>'0');   -- stack level 7 (older)
signal stack_r      : std_logic:='0';                        -- stack read 
signal stack_wb     : std_logic:='0';                        -- stack write prebuf 
signal bank_rp      : std_logic_vector(1 downto 0):="00";             -- memory bank selector
signal bank_irp     : std_logic:='0';                      -- IDF memory bank selector
signal TOn          : std_logic:='1';                      -- Time-out bit
signal PDn          : std_logic:='1';                      -- Power-down bit
signal option_reg   : std_logic_vector(6 downto 0):=(others=>'1');   -- 7 bits because no RBPU bit here
signal intcon_reg   : std_logic_vector(7 downto 0):=(others=>'0');   -- 
signal fsr_reg      : std_logic_vector(7 downto 0):=(others=>'0');   -- 
signal TMR0_reg     : std_logic_vector(7 downto 0):=(others=>'0');   -- 
signal argALU       : std_logic_vector(7 downto 0):=(others=>'0');   -- second arg for ALU, W or bitmask
signal PCLATH       : std_logic_vector(4 downto 0):=(others=>'0');
--signal T0CKI_old    : std_logic:='0';                                -- TOCKI mem for edge detection
signal prescaler    : std_logic_vector(7 downto 0):=(others=>'0');   -- prescaler value
signal presc_event  : std_logic:='0';                        -- ='1' when prescaler should be incremented
signal count_TMR    : std_logic:='0';                        -- ='1' when timer should advance
signal TMR_ovflow   : std_logic:='0';                        -- ='1' when timer overflows
signal reset_ni     : std_logic:='1';                        -- reset signal used internally
--signal WDT          : std_logic_vector(19 downto 0):=(others=>'1');  -- watchdog timer
signal wdt_event    : std_logic:='0';                        -- ='1' when WDT = 0
signal wdt_ri       : std_logic:='0';                        -- ='1' when Istruction = SLEEP or CLRWDT
signal reseti       : std_logic_vector(2 downto 0):="000";   -- counter to provide a 8 cycle reset on WDT timeout
--signal WDTE_buf     : std_logic:='0';                        -- ='1' when Instruction = SLEEP or CLRWDT
signal iskip        : std_logic:='0';                        -- ='1' when instruction requires 2 cycles
signal tskip        : std_logic:='0';                        -- ='1' when instruction is a test
signal IT_sm		  : std_logic_vector(1 downto 0):="00";	 -- state machine for IT branch.
	-- 11 means no IT will be executed
	-- 10 means IT is pending and will be executed (GIE=='1')
	-- 01 means PC should be stacked
	-- 00 means pc should not be stacked
   -- 0x means pc should be loaded with 0x4 and GIE should be zeroed

-- asynchronous signals
signal ALU_output  : std_logic_vector(7 downto 0):=(others=>'0');     -- as name says
signal ALU_coutput : std_logic_vector(8 downto 0):=(others=>'0');     -- as name says, including carry as a 9th bit
signal bitmask     : std_logic_vector(7 downto 0);     -- used for bit operations in ALU
signal write_en    : std_logic:='0';                        -- core write access
signal write_w     : std_logic:='0';                        -- W reg write signal
signal skip        : std_logic:='0';                        -- 1 when next instr is to be skiped
signal local_data  : std_logic_vector(7 downto 0):=(others=>'0');     -- local data bus
signal pending_C   : std_logic:='0';                        -- carry output of ALU
signal pending_Z   : std_logic:='0';                        -- zero output of ALU
signal pending_DC  : std_logic:='0';                        -- DC output of ALU
signal mod_C       : std_logic:='0';                        -- should C be updated ?
signal mod_Z       : std_logic:='0';                        -- should Z be updated ?
signal mod_DC      : std_logic:='0';                        -- should DC be updated ?
signal stack_w     : std_logic:='0';                        -- stack write
signal pending_IT  : std_logic:='0';                        --() there's a pending Interrupt
signal next_pc     : std_logic_vector(10 downto 0):=(others=>'0');     -- address of next instruction
signal next_stpc   : std_logic_vector(10 downto 0):=(others=>'0');     -- address of instruction to eventually stack
signal addr_l      : std_logic_vector(8 downto 0):=(others=>'0');     -- effective data address
signal samplInstr  : std_logic_vector(13 downto 0):=(others=>'0');    -- 
signal ram_select  : std_logic;										-- set to 1 when a transfert is directed to RAM
signal ram_addr    : std_logic_vector(6 downto 0);				-- upper ram @ bits after remap
																					-- bank0 RAM direct mapping
																					-- bank1 RAM mapped on first 32 Bytes

begin


  GIE : process(clk)  -- General Interrupt Enable bit
							 -- bit 7 of INTCON reg (@0xB)
  begin
    if clk'event and clk='1' then
      if reset_ni = '0' then
        intcon_reg(7) <= '0';					-- value at reset
      elsif (write_en = '1') and addr_l(6 downto 0) = "0001011" then
        intcon_reg(7) <= ALU_output(7);	-- memory access
		elsif (IT_sm(1)= '0') then
		  intcon_reg(7) <= '0';					-- interrupt launched
		elsif Instruction="00000000001001" then
		  intcon_reg(7) <= '1';					-- return from interrupt
      end if;
    end if;
  end process GIE;


  ITmgr : process(clk)	-- state machine to launch interrupts
   ------------ IT_sm values ------------------------------------------
  	-- 11 means no IT will be executed
	-- 10 means IT is pending and will be executed (GIE=='1')
	-- 01 means PC should be stacked
	-- 00 means pc should not be stacked
   -- 0x means pc should be loaded with 0x4 and GIE should be zeroed
	--------------------------------------------------------------------
  begin
    if clk'event and clk='1' then
	   if cycle = Dwrite then
		  if IT_sm = "11" then
			 if pending_IT = '1' and intcon_reg(7)='1' then
			   IT_sm <= "10";              -- launch the sequence
		    end if;
		  elsif IT_sm = "10" then
		    if iskip='1' then
			   IT_sm <= "01";
			 elsif tskip = '1' then
			   IT_sm <= "0" & skip;
			 else
			   IT_sm <= "00";					-- stack current or next PC depending of Instruction
			 end if;
		  else
		    IT_sm <= "11";
		  end if;
		end if;
	 end if;
  end process ITmgr;



--  wdbuf : process(clk)
--  begin
--    if clk'event and clk='1' then
--      if reset_n = '0' then
--        WDTE_buf <= WDTE;
--      end if;
--    end if;
--  end process wdbuf;

  TObit : process(clk)
  begin
    if clk'event and clk='1' then
      if reset_n = '0' then
        TOn <= '1';
      elsif wdt_ri = '1' then
        TOn <= '1';
      elsif (wdt_event='1') and (option_reg(3)='0' or presc_event='1') then
        TOn <= '0';
      end if;
    end if;
  end process TObit;


  PDbit : process(clk)
  begin
    if clk'event and clk='1' then
      if reset_n = '0' then
        PDn <= '1';
      elsif Instruction = "00000001100100" then -- CLRWDT instruction
        PDn <= '1';
      elsif cycle = Dsleep then
        PDn <= '0';
      end if;
    end if;
  end process PDbit;



--WDT <= wdt_timing;
wdt_event <= '0';
--  watchdog : process(clk)
--  begin
--    if clk'event and clk='1' then
--      if (reset_ni = '0') or (WDTE_buf='0') or ((wdt_ri = '1') and (cycle=Dread)) then
--        WDT <= wdt_timing;
--        wdt_event <= '0';
--      else
--        WDT <= WDT - 1;
--        if WDT = "00000000000000000000" then
--          wdt_event <= '1';
--        else
--          wdt_event <= '0';
--        end if;
--      end if;
--    end if;
--  end process watchdog;




  wdt_timeout : process(clk)
  begin
    if clk'event and clk='1' then
      if reset_n = '0' then
        reset_ni <= '0';
        reseti   <= "111";        
      elsif (wdt_event='1') and (option_reg(3)='0' or presc_event='1')  and cycle/=Dsleep then
        reset_ni <= '0';
        reseti   <= "111";
      elsif reseti="000" then
        reset_ni <= '1';
        reseti   <= "000";
      else
        reset_ni <= '0';
        reseti   <= reseti - 1;
      end if;
    end if;
  end process wdt_timeout;



  wd_instr : process(clk)
  begin
    if clk'event and clk='1' then
      if (Instruction = "00000001100011") or (Instruction = "00000001100100") then -- SLEEP or CLRWDT
        wdt_ri <= '1';
      else
        wdt_ri <= '0';
      end if;
    end if;
  end process wd_instr;
  
  

  prescaling : process(clk)
  begin
    if clk'event and clk='1' then
      if reset_ni = '0' then
        prescaler <= "11111111";
      elsif (option_reg(3) = '0') and (write_en = '1') and addr_l = "00000001" then
        prescaler <= "11111111";
      elsif (option_reg(3) = '1') and (wdt_ri='1') and (cycle=Dread) then
        prescaler <= "11111111";      
      elsif (count_TMR='1') and (option_reg(3) = '0') then
        prescaler <= prescaler - 1;
      elsif (wdt_event = '1') and (option_reg(3) = '1') then
        prescaler <= prescaler - 1;
      end if;
    end if;
  end process prescaling;
  
  
  
  presc_ev : process(clk) --(prescaler, option_reg(3 downto 0))
  begin
    if clk'event and clk='1' then
      case option_reg(3 downto 0) is
        when          "1000" =>  presc_event <= '1';
        when "0000" | "1001" => if prescaler(0) = '0' then presc_event <= '1'; else presc_event <= '0'; end if;
        when "0001" | "1010" => if prescaler(1 downto 0) = "00" then
                                  presc_event <= '1';
                                else
                                  presc_event <= '0';
                                end if;
        when "0010" | "1011" => if prescaler(2 downto 0) = "000" then
                                  presc_event <= '1';
                                else
                                  presc_event <= '0';
                                end if;
        when "0011" | "1100" => if prescaler(3 downto 0) = "0000" then
                                  presc_event <= '1';
                                else
                                  presc_event <= '0';
                                end if;
        when "0100" | "1101" => if prescaler(4 downto 0) = "00000" then
                                  presc_event <= '1';
                                else
                                  presc_event <= '0';
                                end if;
        when "0101" | "1110" => if prescaler(5 downto 0) = "000000" then
                                  presc_event <= '1';
                                else
                                  presc_event <= '0';
                                end if;
        when "0110" | "1111" => if prescaler(6 downto 0) = "0000000" then
                                  presc_event <= '1';
                                else
                                  presc_event <= '0';
                                end if;
        when others          => if prescaler = "00000000" then
                                  presc_event <= '1';
                                else
                                  presc_event <= '0';
                                end if;       
      end case;
    end if;
  end process presc_ev;



--  sampling_T0in : process(clk)
--  begin
--    if clk'event and clk='1' then
--      T0CKI_old <= T0CKI;
--    end if;
--  end process sampling_T0in;

--T0CKI_old <= '0';

  process(clk)
  begin
    if clk'event and clk='1' then
      if option_reg(5) = '1' then
--        if (option_reg(4)= T0CKI_old) and (T0CKI_old /= T0CKI) then
--          count_TMR <= '1';
--        else 
          count_TMR <= '0';
--        end if;
      else
        if cycle=Dwrite then
          count_TMR <= '1';
        else 
          count_TMR <= '0';
        end if;
      end if;
    end if;
  end process;


  timer : process(clk)    -- processes TMR0 value (timer 0 counter)
								  -- @0x01
  begin
    if clk'event and clk='1' then
      if reset_ni = '0' then
        TMR0_reg <= "00000000";
        TMR_ovflow <= '0';
      elsif (write_en = '1') and addr_l(7 downto 0) = "00000001" then
        TMR0_reg <= ALU_output;
      elsif (count_TMR='1') and (option_reg(3)='1' or presc_event='1') then
        TMR0_reg <= TMR0_reg + 1;
        if TMR0_reg = "11111111" then
          TMR_ovflow <= '1';
        else
          TMR_ovflow <= '0';
        end if;
      elsif cycle=Dinstr then
        TMR_ovflow <= '0';
      end if;
    end if;
  end process timer;
  
  
  
  T0IF : process(clk) -- Timer 0 Interrupt flag bit
							 -- bit 2 of INTCON reg (@0xB)
  begin
    if clk'event and clk='1' then
      if reset_ni = '0' then
        intcon_reg(2) <= '0';					-- default value
      elsif TMR_ovflow = '1' then
        intcon_reg(2) <= '1';					-- set by timer
      elsif (write_en = '1') and addr_l(6 downto 0) = "0001011" then
        intcon_reg(2) <= ALU_output(2);	-- reset by software
      end if;
    end if;
  end process T0IF;


--  INTedge : process(clk)
--  begin
--    if clk'event and clk='1' then
--		prev_INT <= INT;
--    end if;
--  end process INTedge;

--prev_INT <= '0';

  INTF : process(clk) -- INT pin flag bit
							 -- bit 1 of INTCON reg (@0xB)
  begin
    if clk'event and clk='1' then
      if reset_ni = '0' then
        intcon_reg(1) <= '0';						-- default value
--      elsif (option_reg(6) = INT) and (INT /= prev_INT) then
--        intcon_reg(1) <= '1';						-- set by INT event
      elsif (write_en = '1') and addr_l(6 downto 0) = "0001011" then
        intcon_reg(1) <= ALU_output(1);		-- reset by software
      end if;
    end if;
  end process INTF;



intcon_reg(0) <= '0';




  ram_w : process(clk)
  begin
    if clk'event and clk='1' then
      if (write_en = '1') and (ram_select='1') then
        bRAM(to_integer(unsigned(ram_addr))) <= ALU_output;
      end if;
    end if;
  end process ram_w;



  sfr_mgr : process(clk)   -- Specific Function Registers
  begin
    if clk'event and clk='1' then 
      if reset_ni = '0' then
        intcon_reg(6 downto 3) <= "0000";
        option_reg             <= "1111111";
        PCLATH                 <= "00000";
        bank_rp                <= "00";
        bank_irp               <= '0';
        fsr_reg                <= "00000000";
      elsif write_en = '1' then
        case addr_l(6 downto 0) is
          when "0000001" => if addr_l(7) = '1' then
                              option_reg <= ALU_output(6 downto 0);
                            end if;
          when "0000010" => null; -- update pc for further support
          when "0000011" => bank_rp <= ALU_output(6 downto 5); -- update RP bits
									 bank_irp <= ALU_output(7);         -- update iRP bit
          when "0000100" => fsr_reg  <= ALU_output;
          when "0001010" => PCLATH <= ALU_output(4 downto 0);
          when "0001011" => intcon_reg(6 downto 3)  <= ALU_output(6 downto 3);
          when others    => null;
        end case;
      end if;
    end if;
  end process sfr_mgr;


ext_we        <= write_en;
ext_data_out  <= ALU_output;
ext_data_addr <= addr_l;





  process(clk)
  begin
    if clk'event and clk='1' then
      if Instruction(6 downto 0) = "0000000" then
        addr_l <= bank_irp & fsr_reg;
      else
        addr_l <= bank_rp & Instruction(6 downto 0);
      end if;
    end if;
  end process;
  

  process(addr_l(8 downto 4))  -- ram remap
						-- bank0 at legacy addresses
						-- bank1 at first 32 addresses
  begin
   if (addr_l(6 downto 5) = "00") then
		ram_select <= '0';
	elsif (addr_l(8 downto 7) = "00") or (addr_l(6 downto 4)="111") or (addr_l(8 downto 5)="0101") then
		ram_select <= '1';
	else
		ram_select <= '0';
	end if;
  end process;
  
  -- ram remap
		-- bank0 at legacy addresses
		-- bank1 at first 32 addresses
ram_addr(6 downto 5) <= "00" when addr_l(7 downto 5)="101" else addr_l(6 downto 5);  
ram_addr(4 downto 0) <= addr_l(4 downto 0);

  inc_data : process (clk)  -- data produced internally (SFR)
  begin
  if clk'event and clk='1' then
   if Instruction(13) = '0' then
	 if ram_select ='1' then
		local_data <= bRAM(to_integer(unsigned(ram_addr)));
	 else
    case addr_l(6 downto 0) is
      when "0000000" =>     local_data <= "00000000";
      when "0000001" => if addr_l(7)='0' then
                            local_data <= TMR0_reg;
                        else
                            local_data <= '0' & option_reg;
                        end if;
      when "0000010" =>     local_data <= pc(7 downto 0);
      when "0000011" =>     local_data <= bank_irp & bank_rp & TOn & PDn & zero & digcarry & carry;
      when "0000100" =>     local_data <= fsr_reg;
      when "0001010" =>     local_data <= "000" & PCLATH; 
      when "0001011" =>     local_data <= intcon_reg;
      when others    =>     local_data <= ext_data_in;
    end case;
	 end if;
   else
    local_data <= Instruction(7 downto 0);
   end if;
  end if;
  end process inc_data;


  IT_detect : process(clk) 	-- tells if an Interrupt should be taken into account
  begin
    if clk'event and clk='1' then
      if ((intcon_reg(2 downto 0) and intcon_reg(5 downto 3))/="000") then
        pending_IT <= '1';		   -- yes
		else
        pending_IT <= '0';			-- no
      end if;
    end if;
  end process IT_detect;
  


  iskip_mgr : process(clk)		-- manage instructions that skip computing cycles
  begin
    if clk'event and clk='1' then
      if Instruction(13 downto 12) = "10" then  -- GOTO, CALL
        iskip <= '1';
      elsif (Instruction(13 downto 6) = "00000000") and Instruction(3)='1' then  -- RETFIE, RETURN
        iskip <= '1';
      elsif Instruction(13 downto 10) = "1101" then  -- RETLW
        iskip <= '1';
      else
        iskip <= '0';
      end if;
    end if;
  end process iskip_mgr;

  tskip_mgr : process(clk)  -- manage instructions that make test skips
  begin
    if clk'event and clk='1' then
      if (Instruction(13 downto 11)= "001") and (Instruction(9 downto 8)="11") then  --INCFSZ, DECFSZ
        tskip <= '1';
      elsif Instruction(13 downto 11) = "011" then  -- BTFSC, BTFSS
        tskip <= '1';
      else
        tskip <= '0';
      end if;
    end if;
  end process tskip_mgr;



  skip_mgr : process(Instruction, ALU_output, iskip)
					-- should next intsruction be skiped ?
  begin
    if iskip='1' then
      skip <= '1';
    elsif (Instruction(13 downto 11)= "001") and (Instruction(9 downto 8)="11") then  --INCFSZ, DECFSZ
      if ALU_output="00000000" then 
        skip <= '1';
      else
        skip <= '0';
      end if;
    elsif Instruction(13 downto 11) = "011" then  -- BTFSC, BTFSS
      if ALU_output="00000000" then 
        skip <= not Instruction(10);
      else
        skip <= Instruction(10);
      end if;   
    else
      skip <= '0';
    end if;
  end process skip_mgr;


  stackr_mgr : process(clk)
  begin
    if clk'event and clk='1' then
    if cycle /= Dcomp then      
      stack_r <= '0';
    elsif (Instruction(13 downto 6) = "00000000") and (Instruction(3) = '1') then
      stack_r <= '1';
    elsif Instruction(13 downto 10) = "1101" then
      stack_r <= '1';
    else 
      stack_r <= '0';
    end if;
    end if;
  end process stackr_mgr;


  stackwb_mgr : process(clk) 
  begin
  if clk'event and clk='1' then
    if cycle /= Dcomp then 
      stack_wb <= '0';
    elsif Instruction(13 downto 11) = "100" or IT_sm="01" then
      stack_wb <= '1';
	 else
      stack_wb <= '0';
    end if;
  end if;
  end process stackwb_mgr;


  stackw_mgr : process(cycle, IT_sm, skip, stack_wb) 
  begin
    if cycle = Dwrite and It_sm="10" and skip='0' then 
      stack_w <= '1';
    else
      stack_w <= stack_wb;
    end if;
  end process stackw_mgr;

-- stack_w <= stack_wb;


  stack_mgr : process(clk)
  begin
    if clk'event and clk='1' then
      if stack_r='1' and stack_w='0' then
        stack0 <= stack1;
        stack1 <= stack2;
        stack2 <= stack3;
        stack3 <= stack4;
        stack4 <= stack5;
        stack5 <= stack6;
        stack6 <= stack7;
      elsif stack_w='1' and stack_r = '0' then
        stack0 <= next_stpc;
        stack1 <= stack0;
        stack2 <= stack1;
        stack3 <= stack2;
        stack4 <= stack3;
        stack5 <= stack4;
        stack6 <= stack5;
        stack7 <= stack6;
      end if;
    end if;
  end process stack_mgr;

  update_pc : process(clk)
  begin
    if clk'event and clk='1' then
      if reset_ni = '0' then
        pc <= "00000000000";
      elsif cycle = Dwrite then  
        pc <= next_pc;
      end if;  
    end if;
  end process update_pc;



  wnext_pc : process(Instruction, pc, stack_r, stack0, IT_sm)
  begin
    if stack_r = '1' then
      next_pc <= stack0;
    elsif Instruction(13 downto 12) = "10" then
      next_pc <= Instruction(10 downto 0);
    elsif IT_sm(1) = '0' then
	   next_pc <= "00000000100";
	 else
      next_pc <= pc + 1;
    end if;
  end process wnext_pc;


 wnext_stpc : process(Instruction, pc)
  begin        
    if Instruction(13 downto 11)="101" then
      next_stpc <= Instruction(10 downto 0);
    else
      next_stpc <= pc; 
    end if;
  end process wnext_stpc;


  update_status : process(clk)
  begin
    if clk'event and clk='1' then
      if (Instruction(7)='1') and ((instr_code = movwf) or (instr_code=swapf)) and cycle=Dwrite and addr_l(6 downto 0)="0000011" then
        zero <= ALU_output(2);
        carry <= ALU_output(0);
        digcarry <= ALU_output(1);
      else
        if mod_Z = '1' then
          zero <= pending_Z;
        elsif (Instruction(13 downto 11)="010") and (Instruction(9 downto 7)="010") and (addr_l(6 downto 0)="0000011") then
          zero <= Instruction(10);
        end if;
        if mod_C = '1' then
          carry <= pending_C;
        elsif (Instruction(13 downto 11)="010") and (Instruction(9 downto 7)="000") and (addr_l(6 downto 0)="0000011") then
          carry <= Instruction(10);
        end if;
        if mod_DC = '1' then
          digcarry <= pending_DC;
        elsif (Instruction(13 downto 11)="010") and (Instruction(9 downto 7)="001") and (addr_l(6 downto 0)="0000011") then
          digcarry <= Instruction(10);
        end if;
      end if;
    end if;
  end process update_status;


  p_mod_Z : process(Instruction, cycle)
  begin
    if cycle /= Dwrite then 
      mod_Z <= '0';
    elsif Instruction(13 downto 11)="111" then
      mod_Z <= '1';                               -- literal operations
    elsif Instruction(13 downto 8) > "001010" then
      mod_Z <= '0';
    elsif Instruction(13 downto 8) = "000000" then
      mod_Z <= '0';
    else
      mod_Z <= '1';
    end if;
  end process p_mod_Z;
  
  p_mod_C : process(clk)
  begin
  if clk'event and clk='1' then
    if (Instruction(13) =  Instruction(12)) and (cycle = Dcomp) then
      mod_C <= '1';                               -- carry is forwarded when unchanged
    else
      mod_C <= '0';
    end if;
  end if;
  end process p_mod_C;


  p_mod_DC : process(Instruction, cycle)
  begin
    if cycle /= Dwrite then 
      mod_DC <= '0';
    elsif Instruction(13 downto 10)="1111" then
      mod_DC <= '1';                               -- literal operations
    elsif (Instruction(13 downto 11) = "000") and (Instruction(9)='1') and (Instruction(10)=Instruction(8)) then
      mod_DC <= '1';                               -- file ops
    else
      mod_DC <= '0';
    end if;
  end process p_mod_DC;
 
  
  

pending_Z <= '1' when ALU_output = "00000000" else '0';

 
  when_read : process(clk)
  begin
  if clk'event and clk='1' then
    if cycle /= Dinstr then                        -- only during read cycle
      ext_re <= '0';
    elsif Instruction(13) = '1' then              -- no read during branchs and literal ops
      ext_re <= '0';
    elsif Instruction(12 downto 9) = "0000" then  -- no read for clear and misc ops
      ext_re <= '0';
    else                                          -- read for all others
      ext_re <= '1';
    end if;
  end if;
  end process when_read;


  update_w :process(clk)
  begin
    if clk'event and clk='1' then
      if write_w='1' then
        w_reg <= ALU_output;
      end if;
    end if;
  end process update_w;
  


  when_write_en : process(clk)
  begin
  if clk'event and clk='1' then
    if cycle /= Dcomp then
      write_en <= '0';
    elsif Instruction(13) = '1' then
      write_en <= '0';
    elsif Instruction(12) = '1' then
      write_en <= not Instruction(11);       -- Bit operations
    else
      write_en <= Instruction(7);            -- other file operations
    end if;
  end if;
  end process when_write_en;


  when_w_write : process (Instruction, cycle)
  begin
  if cycle /= Dwrite then                           -- write is not possible when data read cycle
    write_w <= '0';
  else
    if Instruction(13 downto 12)="00" then        -- first set of intsructions
      if Instruction(11 downto 8)="0000" then     -- misc instructions ?
        write_w <= '0';
      else
        write_w <= not Instruction(7);            -- instructions with files, send not_d
      end if;
    elsif Instruction (13 downto 12)="11" then    -- instructions with litteral ?
      write_w <= '1';
    else                                          -- other cases : bit tests and gotos
      write_w <= '0';
    end if;
  end if;
  end process when_w_write;


  bit_masquing : process(Instruction(9 downto 7))  -- provides a bit mask considering instruction
  begin
    case Instruction(9 downto 7) is
      when "000"  => bitmask <= "00000001";
      when "001"  => bitmask <= "00000010";
      when "010"  => bitmask <= "00000100";
      when "011"  => bitmask <= "00001000";
      when "100"  => bitmask <= "00010000";
      when "101"  => bitmask <= "00100000";
      when "110"  => bitmask <= "01000000";
      when others => bitmask <= "10000000";
    end case;
  end process bit_masquing;


  wargalu : process(clk)
  begin
    if clk'event and clk='1' then
      if Instruction(13 downto 12) = "01" then
        argALU <= bitmask;
      else
        argALU <= w_reg;
      end if;
    end if;
  end process wargalu;



ALU_output <= ALU_coutput(7 downto 0);
pending_C  <= ALU_coutput(8);

  p_pending_DC : process(instr_code, local_data, w_reg, ALU_output)
  begin
      if instr_code = addn then
        if w_reg(3 downto 0) > ALU_output(3 downto 0) then
          pending_DC <= '1';
        else
          pending_DC <= '0';
        end if;
      else
        if w_reg(3 downto 0) > local_data(3 downto 0) then
          pending_DC <= '0';
        else
          pending_DC <= '1';
        end if;
      end if;
  end process p_pending_DC;


  i_decoding : process(clk)
  begin
   if clk'event and clk='1' then
    if Instruction(13 downto 12) = "00" then
      case Instruction(11 downto 8) is
        when "0000" => instr_code <= movwf;
        when "0001" => instr_code <= clrn;
        when "0010" => instr_code <= subn;
        when "0011" => instr_code <= decn;
        when "0100" => instr_code <= iorn;
        when "0101" => instr_code <= andn;
        when "0110" => instr_code <= xorn;
        when "0111" => instr_code <= addn;
        when "1000" => instr_code <= movn;
        when "1001" => instr_code <= comf;
        when "1010" => instr_code <= incn;
        when "1011" => instr_code <= decn;
        when "1100" => instr_code <= rrf;
        when "1101" => instr_code <= rlf;
        when "1110" => instr_code <= swapf;
        when others => instr_code <= incn;
      end case;
    elsif Instruction(13 downto 12) = "01" then
      case Instruction(11 downto 10) is
        when "00"   =>  instr_code <= bcf;
        when "01"   =>  instr_code <= iorn;
        when others =>  instr_code <= andn;
      end case;
    elsif Instruction(12) = '1' then
      if Instruction(11) = '0' then
        instr_code <= movn;
      elsif Instruction (10) = '1' then
        if Instruction(9) = '1' then
          instr_code <= addn;
        else
          instr_code <= subn;
        end if;
      else
        case Instruction(9 downto 8) is
          when "00"   => instr_code <= iorn;
          when "01"   => instr_code <= andn;
          when "10"   => instr_code <= xorn;
          when others => instr_code <= clrn;  -- unused instruction
        end case;
      end if;
    else
      instr_code <= clrn;                           -- goto, call
    end if;      
   end if;
  end process i_decoding;



  ALU : process(clk)
  begin
    if clk'event and clk='1' then --v0.04
      case instr_code is
        when movwf => ALU_coutput  <= carry & argALU;
        when clrn  => ALU_coutput  <= carry & "00000000";
        when subn  => ALU_coutput  <= ('1' & local_data) - ('0' & argALU);
        when decn  => ALU_coutput  <= carry & (local_data - 1); 
        when iorn  => ALU_coutput  <= carry & (local_data or argALU); 
        when andn  => ALU_coutput  <= carry & (local_data and argALU);
        when xorn  => ALU_coutput  <= carry & (local_data xor argALU);
        when addn  => ALU_coutput  <= ('0' & local_data) + ('0' & argALU);
        when movn  => ALU_coutput  <= carry & local_data;
        when comf  => ALU_coutput  <= carry & (not local_data);
        when incn  => ALU_coutput  <= carry & (local_data + 1);
        when rrf   => ALU_coutput  <= local_data(0) & carry & local_data(7 downto 1);
        when rlf   => ALU_coutput  <= local_data(7 downto 0) & carry;
        when swapf => ALU_coutput  <= carry & local_data(3 downto 0) & local_data(7 downto 4);
        when bcf   => ALU_coutput <= carry & (local_data and (not bitmask));
      end case;
    end if;
  end process ALU;
    
    
    


Instruction <= samplinstr;

  process(clk) -- sample next Instruction
  begin
    if clk'event and clk='1' then
      if reset_ni = '0' then
        samplinstr <= "00000000000000";			-- do NOPs when reset
      elsif cycle = Dwrite then
        if (skip = '1') or (IT_sm /= "11") then
          samplinstr <= "00000000000000";			-- do NOP if skip or IT branch
        else
          samplinstr <= code_data;					-- otherwise, copy code RAM output
        end if;
      end if;
    end if;
  end process;


  cycles : process(clk)                               -- alternate machine cycles
  begin
    if clk'event and clk='1' then
      if reset_ni='0' then             -- start in Dcomp state
        cycle <= Dcomp;
      else
        case cycle is
          when Dinstr => cycle <= Dread;
          when Dread  => cycle <= Dcomp;
          when Dcomp  => if Instruction = "00000001100011" then
                            cycle <= Dsleep;
                         else
                            cycle <= Dwrite;
                         end if;
          when Dwrite => cycle <= Dinstr;
          when Dsleep => if (wdt_event='1') and (option_reg(3)='0' or presc_event='1') then
                            cycle <= Dwrite;
                         elsif pending_IT='1' then
                            cycle <= Dwrite;
                         else
                            cycle <= Dsleep;
                         end if;
        end case;
      end if;
    end if;
  end process cycles;

code_addr <= pc;

END;
